import ast
import base64
import inspect
import json
import logging
import os
import random
import re
import string
import sys

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator, MaxValueValidator
from django.forms import formset_factory
from django.forms import ModelChoiceField
from django.forms.fields import *
from django.template.defaultfilters import filesizeformat
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from django.utils.module_loading import import_string
from django.utils.translation import gettext as _

from . enc import encrypt
from . formsets import build_formset
from . widgets import *
from . settings import *
from . utils import (format_field_name,
                     _split_choices,
                     _split_choices_in_list_canc,
                     _successivo_ad_oggi)

logger = logging.getLogger(__name__)

get_signatures = None
try:
    from filesig.filesig import get_signatures
except Exception as e:
    logger.error('Error: {}'.format(e))

if not get_signatures:
    try:
        from filesig import get_signatures
    except Exception as e:
        logger.error('Error: {}'.format(e))


MAX_UPLOAD_SIZE = getattr(settings, 'MAX_UPLOAD_SIZE', MAX_UPLOAD_SIZE)
ATTACH_NAME_MAX_LEN = getattr(settings, 'ATTACH_NAME_MAX_LEN',
                              ATTACH_NAME_MAX_LEN)
PERMITTED_UPLOAD_FILETYPE = getattr(settings,
                                    'PERMITTED_UPLOAD_FILETYPE',
                                    PERMITTED_UPLOAD_FILETYPE)
WRONG_TYPE = getattr(settings, 'WRONG_TYPE', WRONG_TYPE)
WRONG_SIZE = getattr(settings, 'WRONG_SIZE', WRONG_SIZE)
WRONG_LENGTH = getattr(settings, 'WRONG_LENGTH', WRONG_LENGTH)
IMG_FILETYPE = getattr(settings, 'IMG_FILETYPE', IMG_FILETYPE)
DATA_FILETYPE = getattr(settings, 'DATA_FILETYPE', DATA_FILETYPE)
PDF_FILETYPE = getattr(settings, 'PDF_FILETYPE', PDF_FILETYPE)
P7M_FILETYPE = getattr(settings, 'P7M_FILETYPE', P7M_FILETYPE)
CLASSIFICATION_LIST = getattr(settings, 'CLASSIFICATION_LIST', CLASSIFICATION_LIST)


def get_fields_types(class_name=sys.modules[__name__]):
    fields_types = []
    for m in inspect.getmembers(class_name, inspect.isclass):
        if hasattr(m[1],'field_type'):
            field_type = getattr(m[1], 'field_type')
            if field_type:
                fields_types.append(tuple((m[1].__name__, field_type)))
    fields_types.sort(key=lambda tup: tup[1])
    return fields_types


class BaseCustomField(Field):
    """
    Base Class that defines methods for every field
    """

    # is a complex field
    is_complex = False
    # is a formset
    is_formset = False

    def define_value(self, custom_value=None, **kwargs):
        """
        Integrates fields building with more data coming from user
        configuration params
        """
        return

    def get_fields(self):
        """
        If field isn't complex, it returns a single element list with self.
        Else it returns a child fields list
        """
        return [self]

    def raise_error(self, name, cleaned_data, **kwargs):
        """
        Returns field validation errors list
        """
        return []


class CustomCharField(CharField, BaseCustomField):
    """
    CharField
    """
    field_type = _("Testo")


class CustomEmailField(EmailField, BaseCustomField):
    """
    EmailField
    """
    field_type = _("E-mail")


class CustomURLField(URLField, BaseCustomField):
    """
    URLield
    """
    field_type = _("URL")


class CustomIPField(GenericIPAddressField, BaseCustomField):
    """
    GenericIPAddressField
    """
    field_type = _("Indirizzo IP")

class CustomMACField(CustomCharField, BaseCustomField):
    """
    GenericIPAddressField
    """
    field_type = _("MAC Address")

    def raise_error(self, name, cleaned_data, **kwargs):
        if not cleaned_data: return []

        # Only valid sequences
        if not re.match('([0-9a-fA-F]:?){12}', str(cleaned_data)):
            return [_("Formato non valido"),]

class CustomChoiceField(ChoiceField, BaseCustomField):
    """
    ChoiceField
    """
    def __init__(self, *args, **data_kwargs):
        super().__init__(*args, **data_kwargs)
        self.choices = []

    def define_value(self, choices, **kwargs):
        """
        Set field choices (if present)
        """
        if choices:
            # split string into separate choices
            self.choices += _split_choices(choices)

class CustomMultiChoiceField(MultipleChoiceField, BaseCustomField):
    """
    ChoiceField
    """
    def __init__(self, *args, **data_kwargs):
        super().__init__(*args, **data_kwargs)
        self.choices = []
        self.max_permitted = 0

    def define_value(self, choices, **kwargs):
        """
        Set field choices (if present)
        """
        if choices:
            # if there is a '#' parameter at the end of the string (...#3)
            # defines the maximum number of selectable options
            splitted_string = choices.split("#")
            self.max_permitted = int(splitted_string[1]) if len(splitted_string)>1 else 0
            # split string into separate choices
            self.choices += _split_choices(splitted_string[0])

    def clean(self, *args, **kwargs):
        # args = args if isinstance(args[0], list) else ([args[0]],)
        return super().clean(*args, **kwargs)

    def raise_error(self, name, cleaned_data, **kwargs):
        errors = []
        if cleaned_data and self.max_permitted>0 and len(cleaned_data)>self.max_permitted:
            errors.append(_("Numero massimo di scelte consentito: {}").format(self.max_permitted))
        return errors


class CustomFileField(FileField, BaseCustomField):
    """
    FileField
    """
    field_type = _("Allegato (generico)")

    def __init__(self, *args, **data_kwargs):
        super().__init__(*args, **data_kwargs)

    def raise_error(self, name, cleaned_data, **kwargs):
        data = cleaned_data
        errors = []
        if data:
            msg = ''
            self_valid_extensions = getattr(self, 'valid_extensions', None)
            settings_upload_filetype = PERMITTED_UPLOAD_FILETYPE
            permitted_upload_filetype = self_valid_extensions or settings_upload_filetype
            max_upload_size = MAX_UPLOAD_SIZE
            attach_max_len = ATTACH_NAME_MAX_LEN

            if data.content_type not in permitted_upload_filetype:
                msg_tmpl = WRONG_TYPE
                msg = msg_tmpl.format(permitted_upload_filetype, data.content_type)
            elif data.size > int(max_upload_size):
                msg_tmpl = WRONG_SIZE
                msg = msg_tmpl.format(filesizeformat(max_upload_size),
                                      filesizeformat(data.size))
            # elif len(data._name) > attach_max_len:
            elif len(data.name) > attach_max_len:
                msg_tmpl = WRONG_LENGTH
                # msg = msg_tmpl.format(attach_max_len, len(data._name))
                msg = msg_tmpl.format(attach_max_len, len(data.name))
            if msg: errors.append(msg)
        return errors


class CustomImageField(CustomFileField):
    """
    FileField
    """
    field_type = _("Allegato Immagine")

    def __init__(self, *args, **data_kwargs):
        self.valid_extensions = IMG_FILETYPE
        super().__init__(*args, **data_kwargs)


class CustomDataField(CustomFileField):
    """
    FileField
    """
    field_type = _("Allegato file dati (JSON, CSV, Excel)")

    def __init__(self, *args, **data_kwargs):
        self.valid_extensions = DATA_FILETYPE
        super().__init__(*args, **data_kwargs)


class CustomPDFField(CustomFileField):
    """
    FileField
    """
    field_type = _("Allegato PDF")

    def __init__(self, *args, **data_kwargs):
        self.valid_extensions = PDF_FILETYPE
        super().__init__(*args, **data_kwargs)


class CustomSignedFileField(CustomFileField):
    validation_error = _('Errore di validazione della firma digitale')
    fileformat = ''
    field_type = None

    def get_signature_params(self, data):
        """
        """
        if not data: return
        res = get_signatures(data, type=self.fileformat, only_valids=True)
        return res

    def get_cleaned_signature_params(self, data):
        res = self.get_signature_params(data)
        details = {}
        if res:
            details['Signature Validation'] = res[-1].get('Signature Validation')
            details['Signing Time'] = res[-1].get('Signing Time')
            details['Signer full Distinguished Name'] = res[-1].get('Signer full Distinguished Name')
        return details

    def raise_error(self, name, cleaned_data, **kwargs):
        data = cleaned_data
        errors = []
        if data:
            simple_filefield_errors = super().raise_error(name, cleaned_data, **kwargs)
            errors.extend(simple_filefield_errors)
            res = self.get_signature_params(data)
            if not res or 'is valid' not in res[-1].get('Signature Validation').lower():
                errors.append(self.validation_error)
        return errors


class CustomSignedPdfField(CustomSignedFileField):
    """
    """
    field_type = _("Allegato PDF firmato")
    fileformat = 'pdf'
    valid_extensions = PDF_FILETYPE


class CustomSignedP7MField(CustomSignedFileField):
    """
    """
    field_type = _("Allegato P7M firmato")
    fileformat = 'p7m'
    valid_extensions = P7M_FILETYPE


class PositiveIntegerField(DecimalField, BaseCustomField):
    """
    Positive integer DecimalField
    """
    field_type = _("Numero intero positivo")
    default_validators = [MinValueValidator(0)]

    def __init__(self, *args, **data_kwargs):
        # Non si accettano formati con cifre decimali
        data_kwargs['decimal_places'] = 0
        super().__init__(*args, **data_kwargs)

    def raise_error(self, name, cleaned_data, **kwargs):
        if not cleaned_data: return []
        # Only numbers (expressions like 16e50 aren't permitted)
        if not re.match('^[0-9]+$', str(cleaned_data)):
            return [_("Solo numeri ammessi"),]


class PositiveFloatField(DecimalField, BaseCustomField):
    """
    Positive float DecimalField
    """
    field_type = _("Numero con virgola positivo")
    default_validators = [MinValueValidator(0)]

    def __init__(self, *args, **data_kwargs):
        # Max 3 cifre decimali
        data_kwargs['decimal_places'] = 3
        super().__init__(*args, **data_kwargs)

    def raise_error(self, name, cleaned_data, **kwargs):
        if not cleaned_data: return []
        # Only numbers (expressions like 16e50 aren't permitted)
        if not re.match('^[0-9]+\.?[0-9]?$', str(cleaned_data)):
            return [_("Solo numeri ammessi"),]


class TextAreaField(CharField, BaseCustomField):
    """
    TextArea
    """
    field_type = _("Testo lungo")
    widget = forms.Textarea


class CheckBoxField(BooleanField, BaseCustomField):
    """
    BooleanField Checkbox
    """
    field_type = _("Checkbox")
    widget = forms.CheckboxInput

class MultiCheckBoxField(CustomMultiChoiceField):
    """
    BooleanField Checkbox multi-value
    """
    field_type = _("Checkbox multi-valore")
    widget = forms.CheckboxSelectMultiple


class CustomSelectBoxField(CustomChoiceField):
    """
    SelectBox
    """
    field_type = _("Lista di opzioni (tendina)")

    def __init__(self, *args, **data_kwargs):
        super().__init__(*args, **data_kwargs)
        self.choices = [('', _('Scegli una opzione')),]


class CustomRadioBoxField(CustomChoiceField):
    """
    RadioBox
    """
    field_type = _("Lista di opzioni (checkbox)")
    widget = forms.RadioSelect


class BaseDateField(DateField, BaseCustomField):
    """
    DateField
    """
    field_type = _("Data")
    input_formats = settings.DATE_INPUT_FORMATS


class BaseDateTimeSimpleField(DateTimeField, BaseCustomField):
    """
    DateTimeField
    """
    field_type = _("Data e ora (campo singolo)")
    input_formats = settings.DATETIME_INPUT_FORMATS


class BaseDateTimeField(BaseCustomField):
    """
    DateTimeField
    """
    field_type = _("Data e Ora (campi separati)")
    is_complex = True

    def __init__(self, *args, **data_kwargs):
        # Date DateField
        self.data = BaseDateField(*args, **data_kwargs)
        self.data.label = _("{} (Data)").format(data_kwargs.get('label'))
        self.data.name = "{}_dyn".format(format_field_name(self.data.label))
        self.data.parent = self

        # Hour SelectBox
        self.hour = CustomSelectBoxField(*args, **data_kwargs)
        self.hour.label = _("{} (Ore)").format(data_kwargs.get('label'))
        self.hour.name = "{}_dyn".format(format_field_name(self.hour.label))
        self.hour.choices = [(i,i) for i in range(24)]
        self.hour.initial = 0
        self.hour.parent = self

        # Minutes SelectBox
        self.minute = CustomSelectBoxField(*args, **data_kwargs)
        self.minute.label = _("{} (Minuti)").format(data_kwargs.get('label'))
        self.minute.name = "{}_dyn".format(format_field_name(self.minute.label))
        self.minute.choices = [(i,i) for i in range(60)]
        self.minute.initial = 0
        self.minute.parent = self

    def get_fields(self):
        return [self.data, self.hour, self.minute]


class DateStartEndComplexField(BaseCustomField):
    """
    Field composed by StartDate (DateField) and EndDate (DateField)
    """
    field_type = _("Data inizio e Data fine")
    is_complex = True

    def __init__(self, *args, **data_kwargs):
        parent_label = data_kwargs.get('label')

        # Start date
        self.start = BaseDateField(*args, **data_kwargs)
        self.start.required = data_kwargs.get('required')
        self.start.label = _("{} (Data inizio)").format(parent_label)
        self.start.name = "{}_dyn".format(format_field_name(self.start.label))

        # Set child parent
        self.start.parent = self

        # End date
        self.end = BaseDateField(*args, **data_kwargs)
        self.end.required = data_kwargs.get('required')
        self.end.label = _("{} (Data fine)").format(parent_label)
        self.end.name = "{}_dyn".format(format_field_name(self.end.label))

        # Set child parent
        self.end.parent = self

    def get_fields(self):
        fields = [self.start, self.end]
        return fields

    def raise_error(self, name, cleaned_data, **kwargs):
        errors = []
        start_value = cleaned_data.get(self.start.name)
        end_value = cleaned_data.get(self.end.name)

        # If inizialization fails
        if not cleaned_data.get(name): return []

        # Check Start date
        if name == self.start.name:
            # if start_date > end_date
            if end_value and start_value > end_value:
                errors.append(_("La data di inizio non può "
                                "essere successiva a quella di fine"))

        return errors


class ProtocolloField(BaseCustomField):
    """
    Protocolo type, number and date (or another classification type)
    """
    field_type = "Protocollo (tipo/numero/data)"
    is_complex = True

    def __init__(self, *args, **data_kwargs):
        parent_label = data_kwargs.get('label')

        # Procotol type (selectbox)
        self.tipo = CustomSelectBoxField(*args, **data_kwargs)
        self.tipo.required = data_kwargs.get('required')
        self.tipo.label = _("{} (Tipo numerazione)").format(parent_label)
        self.tipo.name = "{}_dyn".format(format_field_name(self.tipo.label))
        self.tipo.help_text = _("Scegli se protocollo/decreto/delibera, "
                                "al/alla quale la numerazione è riferita")
        self.tipo.choices += [(i[0].lower().replace(' ', '_'), i[1]) \
                             for i in CLASSIFICATION_LIST]
        self.tipo.parent = self

        # Protocol number (char field)
        self.numero = CustomCharField(*args, **data_kwargs)
        self.numero.required = data_kwargs.get('required')
        self.numero.label = _("{} (Numero Protocollo/Delibera/Decreto)").format(parent_label)
        self.numero.name = "{}_dyn".format(format_field_name(self.numero.label))
        self.numero.help_text = _("Indica il numero del "
                                  "protocollo/decreto/delibera")
        self.numero.parent = self

        # Protocol date (DateField)
        self.data = BaseDateField(*args, **data_kwargs)
        self.data.required = data_kwargs.get('required')
        self.data.label = _("{} (Data Protocollo/Delibera/Decreto)").format(parent_label)
        self.data.name = "{}_dyn".format(format_field_name(self.data.label))
        self.data.help_text = _("Indica la data del protocollo/delibera/decreto")
        self.data.parent = self

    def get_fields(self):
        return [self.tipo, self.numero, self.data]

    def get_data_name(self):
        return self.data.name

    def raise_error(self, name, cleaned_data, **kwargs):
        value = cleaned_data.get(name)

        if not value: return [_("Valore mancante")]

        # Check protocol date
        if name == self.data.name:
            errors = []
            # If protocol_date > today
            if _successivo_ad_oggi(value):
                errors.append(_("La data di protocollo non può essere"
                                " successiva ad oggi"))
            return errors


class CustomHiddenField(CharField, BaseCustomField):
    """
    CharField Hidden
    """
    field_type = _("Campo nascosto")
    widget = forms.HiddenInput

    def define_value(self, custom_value, **kwargs):
        self.widget = forms.HiddenInput(attrs={'value': custom_value})


class CaptchaHiddenField(CustomHiddenField):
    """
    Captcha Hidden field
    """
    field_type = ""


class CaptchaField(BaseCustomField):
    """
    Captcha

    in settings:
        CAPTCHA_FONTS = ['/usr/share/fonts/truetype/ttf-bitstream-vera/Vera.ttf',
                         '/usr/share/fonts/truetype/liberation/LiberationMono-Italic.ttf']
        CAPTCHA_SECRET = b'6sa78d6as83$_RDF'
        CAPTCHA_SALT = b'ingoalla'

    """
    widget = CaptchaWidget

    def define_value(self, custom_value, **kwargs):
        self.widget = CaptchaWidget(attrs={'value': custom_value,
                                           'hidden_field': kwargs.get('hidden_field', ''),
                                           'lang': kwargs.get('lang', getattr(settings,
                                                                              'CAPTCHA_DEFAULT_LANG',
                                                                              CAPTCHA_DEFAULT_LANG))})


class CustomCaptchaComplexField(BaseCustomField):
    """
    Captcha complex field (image, sound, input and hidden text)
    """
    field_type = _("Captcha")
    is_complex = True

    def __init__(self, *args, **kwargs):
        # for auto-generated CaPTCHA
        # you can pass in kwargs captcha_name and captcha_hidden_name
        captcha_name = kwargs.pop("captcha_name") if kwargs.get("captcha_name") else ""
        captcha_hidden_name = kwargs.pop("captcha_hidden_name") if kwargs.get("captcha_hidden_name") else ""

        # CaPTCHA
        parent_label = kwargs.get('label')
        length = getattr(settings, 'CAPTCHA_LENGTH', 5)
        self.text = ''.join([random.choice(string.ascii_letters) for i in range(length)]) # nosec

        self.hidden_data = '{"created": "' + timezone.localtime().isoformat() + '", "text": "' + self.text + '"}'

        # self.value = base64.b64encode(encrypt(self.text)).decode()
        self.value = base64.b64encode(encrypt(self.hidden_data)).decode()

        # self.captcha_hidden.define_value(custom_value=value)
        logger.debug(self.text, self.value)

        # Captcha hidden field
        self.captcha_hidden = CaptchaHiddenField()
        self.captcha_hidden.required = True
        self.captcha_hidden.name = captcha_hidden_name or "{}_hidden_dyn".format(format_field_name(parent_label))
        self.captcha_hidden.label = ''
        self.captcha_hidden.parent = self

        # Captcha field
        self.captcha = CaptchaField(*args, **kwargs)
        self.captcha.required = True
        # self.captcha.define_value(custom_value=text,
                                  # hidden_field="id_{}".format(self.captcha_hidden.name),
                                  # lang=self.lang)
        self.captcha.label = parent_label
        self.captcha.name = captcha_name or "{}_dyn".format(format_field_name(parent_label))
        # this should be taken from constructore dict
        # self.captcha.help_text = _("CaPTCHA: inserisci i caratteri/numeri raffigurati nell'immagine")
        self.captcha.parent = self

    def define_value(self, custom_value, **kwargs):
        # captcha_hidden field define_value
        self.captcha_hidden.define_value(custom_value=self.value)
        # get language from form custom_params
        # default 'en' (see captcha.py)
        lang = kwargs.get('lang', getattr(settings,
                                          'CAPTCHA_DEFAULT_LANG',
                                          CAPTCHA_DEFAULT_LANG))
        # captcha field define_value (with language code)
        self.captcha.define_value(custom_value=self.text,
                                  hidden_field="id_{}".format(self.captcha_hidden.name),
                                  lang=lang)

    def get_fields(self):
        return [self.captcha, self.captcha_hidden]

    def raise_error(self, name, cleaned_data, **kwargs):
        """
        """
        _err_msg = _('CaPTCHA not valid!')
        errors = []
        value = cleaned_data.get(self.captcha.name)
        check = cleaned_data.get(self.captcha_hidden.name)
        try:
            cvalue = decrypt(check).decode()
            json_check = json.loads(cvalue)
            created = json_check['created']
            clear_text = json_check['text']

            # if a value is submitted raise errors
            # else required field error is raised
            if value:
                created_date = parse_datetime(created)
                expiration_time = getattr(settings, 'CAPTCHA_EXPIRATION_TIME', CAPTCHA_EXPIRATION_TIME)
                is_expired = timezone.localtime() > (created_date + timezone.timedelta(milliseconds=int(expiration_time)))

                # if captcha is expired
                # check on name
                # (prevent to call 2 times this control, one for each child field)
                if name == self.captcha.name and is_expired:
                    errors.append(_('CaPTCHA expired'))

                # if captcha is incorrect
                # if value and cvalue.lower() != value.lower():
                elif clear_text.lower() != value.lower():
                    errors.append(_err_msg)
        except Exception as e:
            errors.append(e)

        return errors


class CustomComplexTableField(ChoiceField, BaseCustomField):
    """
    Formset field
    """
    # error message if is_required and no forms are present
    validation_error = _("Questo campo necessita di almeno una riga")
    field_type = _("Inserimenti multipli")
    is_complex = True
    is_formset = True
    choices = None

    def get_formset(self):
        if not self.choices: return
        return build_formset(choices=self.choices)

    def clean(self, *args, **kwargs):
        # If is_required and there aren't forms raise ValidationError.
        # Else, each form will clean itself, raising its own fields errors.
        if not self.widget.formset:
            raise ValidationError(self.validation_error)
        if self.required and not self.widget.formset.forms:
            raise ValidationError(self.validation_error)
        return

    def define_value(self, custom_value, **kwargs):
        # Set field choices (if 'custom_value')
        if custom_value:
            elements = _split_choices_in_list_canc(custom_value)
            self.choices = elements



from django.utils.translation import ugettext_lazy as _


CLASSIFICATION_LIST = (('protocollo', _('Protocollo')),
                       ('decreto_rettorale', _('Decreto Rettorale (D.R.)')),
                       ('decreto_direttore_generale', _('Decreto del Direttore Generale (D.D.G.)')),
                       ('decreto_dirigente_struttura', _('Decreto del Direttore Dipartimento o Dirigente Struttura')),
                       ('decreto_direttore_cr', _('Decreto del Direttore del Centro Residenziale (D.CR.)')),
                       ('decreto_prorettore', _('Decreto del Prorettore (Centro Residenziale)')),
                       ('delibera_dipartimento_facolta', _('Delibera di Dipartimento/Facoltà')),
                       ('delibera_senato', _('Delibera del Senato')),
                       ('delibera_cda', _('Delibera del C.D.A.')))

# 2.5MB - 2621440
# 5MB - 5242880
# 10MB - 10485760
# 20MB - 20971520
# 50MB - 5242880
# 100MB 104857600
# 250MB - 214958080
# 500MB - 429916160
MAX_UPLOAD_SIZE = 10485760

PDF_FILETYPE = ('application/pdf',)
DATA_FILETYPE = ('text/csv', 'application/json',
                 'application/vnd.ms-excel',
                 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
                 'application/vnd.oasis.opendocument.spreadsheet',
                 'application/wps-office.xls',
                 )
TEXT_FILETYPE = ('text/plain',
                 'application/vnd.oasis.opendocument.text',
                 'application/msword',
                 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
                )
IMG_FILETYPE = ('image/jpeg', 'image/png', 'image/gif', 'image/x-ms-bmp')
P7M_FILETYPE = ('application/pkcs7-mime',)
SIGNED_FILETYPE = PDF_FILETYPE + P7M_FILETYPE
PERMITTED_UPLOAD_FILETYPE = TEXT_FILETYPE + DATA_FILETYPE + SIGNED_FILETYPE + IMG_FILETYPE

# maximum permitted filename lengh in attachments, uploads
ATTACH_NAME_MAX_LEN = 50

ATTACHMENTS_DICT_PREFIX = "allegati"

# attachments validation messages
WRONG_TYPE = _("Per favore esegui l'upload di soli file "
               "in {}. "
               "Attualmente questo è '{}'")
WRONG_SIZE = _("Per favore mantieni la dimensione del file entro {}. "
               "Attualmente questo è {}")
WRONG_LENGTH = _("Per favore usa una lunghezza massima del nome dell'allegato "
                 "inferiore a {}. Attualmente hai inserito un nome di {} caratteri")

# formset special words
FORMSET_REGEX = "^(?P<field_name>{})-(?P<index>[0-9]+)-(?P<name>[a-zA-ZÀ-ÿ0-9_°^\-\'\"]+)$"
FORMSET_FULL_REGEX = FORMSET_REGEX.format("[a-zA-ZÀ-ÿ0-9_°^\-\'\"]+")

FORMSET_TEMPLATE_NAMEID = 'NNNNN'
MANAGEMENT_FORMSET_STRINGS = [FORMSET_TEMPLATE_NAMEID,
                              '-TOTAL_FORMS', '-INITIAL_FORMS',
                              '-MAX_NUM_FORMS','-MIN_NUM_FORMS']

CAPTCHA_DEFAULT_LANG = 'en'
CAPTCHA_EXPIRATION_TIME = 120000 # milliseconds
CUSTOM_WIDGETS_IN_FORMSETS = True



import os

from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase

from django_form_builder import dynamic_fields
from django_form_builder.forms import BaseDynamicForm
from django_form_builder.models import *
from django_form_builder.utils import format_field_name


class BaseTest(TestCase):

    def get_baseform(self, initial_fields, data={}, files={}):
        """
        Build a form from BaseDynamicForm
        """
        form = BaseDynamicForm.get_form(initial_fields=initial_fields,
                                        data=data,
                                        files=files)
        return form

    def create_field(self, field_class, field_label='', required=True):
        """
        Create a field using class name
        """
        label = field_label or 'test field'
        field_data = {'required': required,
                      'label': label}
        field = getattr(dynamic_fields, field_class)(**field_data)
        if field.is_complex:
            return field.get_fields()
        return [field,]

    def create_fake_file(self, name="test", ext="pdf",
                         content_type="application/pdf"):
        """
        Create a fake file to simulate upload
        """
        return SimpleUploadedFile("{}.{}".format(name, ext),
                                  b"file_content",
                                  content_type=content_type)

    def get_file(self, real_file_name,
                 name="test", ext="pdf",
                 content_type="application/pdf"):
        """
        Simulate upload using a real file
        """
        if not real_file_name: return None
        module_dir = "{}/{}".format(os.path.dirname(__file__), "files")
        file_path = os.path.join(module_dir, real_file_name)
        file_obj = open(file_path, 'rb')
        return SimpleUploadedFile("{}.{}".format(name, ext),
                                  file_obj.read(),
                                  content_type=content_type)

    def single_field_form(self, field_class,
                          data_value='{}', file_value='{}',
                          required=True):
        """
        Build a form with a single field
        """
        field = self.create_field(field_class=field_class,
                                  required=required)[0]
        field_id = format_field_name(field.label)
        initial_fields = {field_id: field}
        data = {field_id: data_value}
        files = {field_id: file_value}
        form = self.get_baseform(initial_fields=initial_fields,
                                 data=data,
                                 files=files)
        return form

    def complex_field_form(self, field_class, data, field_label=''):
        """
        Build a form with a single complex field
        """
        fields = self.create_field(field_class, field_label)
        initial_fields = {}
        for field in fields:
            initial_fields[field.name] = field
        data = data
        form = self.get_baseform(initial_fields=initial_fields,
                                 data=data)
        return form

    def get_filepath(self):
        return "{}/{}".format(os.path.dirname(__file__), "files")

    def get_full_filepath(self, filename):
        module_dir = "{}/{}".format(os.path.dirname(__file__), "files")
        return os.path.join(module_dir, filename)



import json
import re

from django.conf import settings
from django.utils import timezone
from django.utils.html import strip_tags

from . settings import (ATTACHMENTS_DICT_PREFIX,
                        MANAGEMENT_FORMSET_STRINGS)


ATTACHMENTS_DICT_PREFIX = getattr(settings, 'ATTACHMENTS_DICT_PREFIX',
                                  ATTACHMENTS_DICT_PREFIX)
MANAGEMENT_FORMSET_STRINGS = getattr(settings, 'MANAGEMENT_FORMSET_STRINGS',
                                     MANAGEMENT_FORMSET_STRINGS)


def format_field_name(field_name, lower=True):
    f = field_name.replace(' ','_')
    f = f.replace('(','')
    f = f.replace(')','')
    f = f.replace('/','_')
    if lower: return f.lower()
    return f.upper


def get_labeled_errors(form):
    d = {}
    for field_name in form.errors:
        field = form.fields[field_name]
        d[field.label] = form.errors[field_name]
    return d


def get_formset_labeled_errors(formset_errors):
    d = {}
    for err in formset_errors:
        for k,v in err.items():
            d[k] = strip_tags(v)
    return d


def _split_choices(choices_string):
    """
    Riceve una stringa e la splitta ogni ';'
    creando una tupla di scelte
    """
    str_split = choices_string.split(';')
    choices = tuple((x, x) for x in str_split)
    return choices


def _split_choices_in_list(choices_string):
    """
    Riceve una stringa e la splitta ogni ';'
    creando una lista di scelte
    """
    str_split = choices_string.split(';')
    return str_split


def _split_choices_in_list_canc(choices_string):
    """
    Riceve una stringa e la splitta ogni '#'
    creando una lista di scelte
    """
    str_split = choices_string.split('#')
    return str_split


def _successivo_ad_oggi(data_da_valutare):
    """
    Ci dice se una data è successiva alla data di oggi
    """
    oggi = timezone.localdate()
    if data_da_valutare:
        return data_da_valutare > oggi


# Unused formset method
def get_formset_dict(data, clean=False):
    """data: dict

    returns for example:
        {'date_data_dyn': '2019-05-10',
         'date_minuti_dyn': '25',
         'date_ore_dyn': '5',
         'formsets': {'form-0': {'ccc': 'dfgdfgdfg',
           'co': 'sono',
           'col1': 'un',
           'data': '2019-04-03'},
          'form-1': {'ccc': '234234',
           'co': 'son44o',
           'col1': 'un333',
           'data': '2016-04-03'}},
         'ticket_description': 'dfgdfg',
         'ticket_subject': 'fgdfg'}
    """
    # crea la lista se nei campi esistono formsets
    _re = '(?P<prefix_type>.*)-(?P<prefix_index>[0-9]+)-(?P<name>.*)'
    fields_to_be_removed = []
    d = data.copy()

    formset_dict = {}

    formset_dict_catalog = {}
    for i in d:
        r =  re.match(_re, i)
        if not r: continue
        r = r.groupdict()
        formset_name = '{}-{}'.format(r['prefix_type'],
                                      r['prefix_index'])
        if not formset_dict.get(formset_name):
            formset_dict[formset_name] = {}
        formset_dict[formset_name][r['name']] = d[i]
        fields_to_be_removed.append(i)

    # clean up
    if clean:
        for i in fields_to_be_removed:
            d.pop(i)

    d['formsets'] = formset_dict
    return d


def get_formset_list(data):
    """
    return a list like
    [{'form-0-ccc': 'dfgdfgdfg',
      'form-0-co': 'sono',
      'form-0-col1': 'un',
      'form-0-data': '2019-04-03'},
     {'form-1-ccc': '234234',
      'form-1-co': 'son44o',
      'form-1-col1': 'un333',
      'form-1-data': '2016-04-03'}]
    """
    d = get_formset_dict(data, clean=True)
    l = []
    for i in d.get('formsets'):
        formd = dict()
        for e in d['formsets'][i]:
            k = e #'{}-{}'.format(i,e)
            v = d['formsets'][i][e]
            formd[k] = v
        l.append(formd)
    return l


def get_as_dict(compiled_module_json={},
                allegati=True, formset_management=True):
    """
    torna il dizionario con gli allegati raggruppati in 'allegati'
    """
    # d = json.loads(self.modulo_compilato)
    if compiled_module_json.get(ATTACHMENTS_DICT_PREFIX):
        if not allegati: del compiled_module_json[ATTACHMENTS_DICT_PREFIX]

    to_be_removed = []
    # Corregge i campi inseriti con spazi prima e dopo
    for k,v in compiled_module_json.items():
        # Se non vogliamo che si porti dietro
        # i field input del management del formset (vedi parametri)
        # crea una lista di to_be_removed
        if not formset_management:
            # Nel settings definiamo le parole chiave
            # per identificare i field di management
            for word in MANAGEMENT_FORMSET_STRINGS:
                if word in k:
                    to_be_removed.append(k)
                    break

        if isinstance(compiled_module_json[k],str):
            compiled_module_json[k] = compiled_module_json[k].strip()

    # Se 'formset_management=False', elimina dal dict i to_be_removed
    if not formset_management:
        for r in to_be_removed:
            del compiled_module_json[r]

    return compiled_module_json

def get_allegati_dict(compiled_module, pure_text=False):
    json_dict = compiled_module if pure_text else json.loads(compiled_module.modulo_compilato)
    allegati_dict = get_as_dict(json_dict).get(ATTACHMENTS_DICT_PREFIX)
    if allegati_dict:
        return allegati_dict
    else:
        return {}

def get_allegati(compiled_module):
    """
    torna un lista di file path contenenti gli allegati su filesystem
    """
    allegati = get_allegati_dict(compiled_module)
    if allegati:
        return allegati.items()
    else:
        return []

def get_as_dict_with_allegati(compiled_module, pure_text=False):
    """
    torna il dizionario senza la chiave ATTACHMENTS_DICT_PREFIX
    per gli allegati
    """
    json_dict = compiled_module if pure_text else json.loads(compiled_module.modulo_compilato)
    d = get_as_dict(json_dict)

    # ricostruisco la struttura con gli allegati
    allegati = d.get(ATTACHMENTS_DICT_PREFIX)
    if allegati:
        del d[ATTACHMENTS_DICT_PREFIX]
        d.update(allegati)
    return d

def set_as_dict(obj, modulo_compilato_dict, fields_to_pop=[], indent=False):
    for field in fields_to_pop:
        modulo_compilato_dict.pop(field)
    obj.modulo_compilato = json.dumps(modulo_compilato_dict) \
                           if not indent \
                           else json.dumps(modulo_compilato_dict, indent=2)
    obj.save()

def get_POST_as_json(request, fields_to_pop=[], indent=False):
    data = {}
    for k in request.POST:
        value_list = request.POST.getlist(k)
        if len(value_list) > 1:
            data[k] = value_list
        else:
            data[k] = request.POST[k]
    # data = {k:v for k,v in request.POST.items()}
    for field_name in fields_to_pop:
        if field_name in data:
            data.pop(field_name)
    if "csrfmiddlewaretoken" in data: data.pop("csrfmiddlewaretoken")
    json_data = json.dumps(data) if not indent else json.dumps(data, indent=2)
    return json_data



import inspect

from django import template

from django_form_builder import dynamic_fields

register = template.Library()

@register.filter
def get_dyn_field_name(value):
    for m in inspect.getmembers(dynamic_fields, inspect.isclass):
        if m[0]==value: return getattr(m[1], 'field_type')
    return value

@register.simple_tag
def get_attachment_sign_details(form, field_path, field_name, field_value):
    field = form.fields.get(field_name, None)
    if field and isinstance(field, dynamic_fields.CustomSignedFileField):
        return field.get_cleaned_signature_params('{}/{}'.format(field_path, field_value))
    return False



import ast
import re
import sys
from django import forms
from django.conf import settings
from django.utils.module_loading import import_string

from . settings import CUSTOM_WIDGETS_IN_FORMSETS
from . utils import format_field_name, _split_choices


def get_empty_form(form_class=forms.Form):
    class Dynamic(form_class):
        pass
    return Dynamic


def build_formset(choices, extra=0, required=False, prefix='form', data={}, files={}):
    """ Get formset
    """
    _regexp = '(?P<colname>[a-zA-Z0-9À-ÿ\'\"\_ ]*)\((?P<coldict>[\{\}\.0-9a-zA-ZÀ-ÿ\'\"\:\;\_\,\s\- ]*)\)'
    min_num = 0
    if required: min_num = 1
    eform = get_empty_form()
    for choice in choices:
        colname = format_field_name(choice) # needed for simple CharField withoud attrs
        contenuto = re.search(_regexp, choice)
        field_dict = None
        mod_name = __package__ + '.dynamic_fields'
        sysmod = sys.modules[mod_name]
        if contenuto:
            coldict = contenuto.groupdict().get('coldict')
            colname = format_field_name(contenuto.groupdict()['colname'])
            if coldict:
                field_dict = ast.literal_eval(coldict)
                field_type_name = field_dict['type']
                del field_dict['type']
                custom_field = getattr(sysmod, field_type_name)(**field_dict) \
                               if hasattr(sysmod, field_type_name) \
                               else getattr(sysmod, 'CustomCharField')()
                # choice if use or not fields custom widget
                # javascript may cause some problem
                use_custom_widget = getattr(settings, 'CUSTOM_WIDGETS_IN_FORMSETS', CUSTOM_WIDGETS_IN_FORMSETS)
                if use_custom_widget:
                    custom_widget = getattr(settings, 'CUSTOM_WIDGETS').get(field_type_name) \
                                    if hasattr(settings, 'CUSTOM_WIDGETS') else None
                    if custom_widget:
                        custom_field.widget = import_string(custom_widget)()

                if field_dict.get('choices'):
                    custom_field.choices += _split_choices(field_dict.get('choices'))
        else:
            # custom_field = forms.CharField()
            custom_field = getattr(sysmod, 'CustomCharField')()
        eform.base_fields[colname] = custom_field
        eform.declared_fields[colname] = custom_field

    # Django formset
    fac =  forms.formset_factory(eform, extra=extra, min_num=min_num)
    if data:
        return fac(prefix=prefix, data=data, files=files)
    return fac(prefix=prefix)



import logging

from django.test import TestCase

from django_form_builder import dynamic_fields
from django_form_builder.forms import BaseDynamicForm
from django_form_builder.models import *
from django_form_builder.utils import (_split_choices_in_list,
                                       _split_choices_in_list_canc,
                                       get_formset_dict,
                                       get_formset_list,
                                       get_formset_labeled_errors,
                                       get_labeled_errors)
from django_form_builder.widgets import FormsetdWidget

from . base import BaseTest


logger = logging.getLogger('my_logger')

class TestFormsets(BaseTest):

    def test_formset(self):
        """
        Build a formset passing valid example data
        """
        choices_string = "col1({'type':'CustomSelectBoxField', 'choices':'v1;v2;v3'})#"\
                         "co({'type':'BaseDateField'})#"\
                         "ccc"
        prefix = "prefix"
        data = {
            '{}-TOTAL_FORMS'.format(prefix): '1',
            '{}-INITIAL_FORMS'.format(prefix): '0',
            '{}-MAX_NUM_FORMS'.format(prefix): '',
            '{}-0-col1'.format(prefix): 'v1',
            '{}-0-co'.format(prefix): '1904-06-16',
            '{}-0-ccc'.format(prefix): 'Plain text',
        }

        choices = _split_choices_in_list_canc(choices_string)
        formset = dynamic_fields.build_formset(choices=choices,
                                               required=True,
                                               prefix=prefix,
                                               data=data)
        logger.info("Test form with formset")
        assert formset.is_valid()

    def test_formset_invalid(self):
        """
        Build a formset passing invalid example data
        """
        choices_string = "col1({'type':'CustomSelectBoxField', 'choices':'v1;v2;v3'})#"\
                         "co({'type':'BaseDateField'})#"\
                         "ccc"
        prefix = "prefix"
        data = {
            '{}-TOTAL_FORMS'.format(prefix): '1',
            '{}-INITIAL_FORMS'.format(prefix): '0',
            '{}-MAX_NUM_FORMS'.format(prefix): '',
            '{}-0-col1'.format(prefix): 'v1',
            '{}-0-co'.format(prefix): '1904-106-16',
            '{}-0-ccc'.format(prefix): 'Plain text',
        }

        choices = _split_choices_in_list_canc(choices_string)
        formset = dynamic_fields.build_formset(choices=choices,
                                               required=True,
                                               prefix=prefix,
                                               data=data)
        logger.info("Test form with formset (invalid)")
        logger.info(get_formset_labeled_errors(formset.errors))
        self.assertFalse(formset.is_valid())

    def test_split_choices_in_list(self):
        """
        Test splitting choices string
        """
        choices = 'v1;v2;v3'
        logger.info("Test _split_choices_in_list()")
        assert _split_choices_in_list(choices)

    def test_formset_list(self):
        """
        Test get_formset_list() method
        """
        prefix = "prefix"
        data = {
            '{}-TOTAL_FORMS'.format(prefix): '1',
            '{}-INITIAL_FORMS'.format(prefix): '0',
            '{}-MAX_NUM_FORMS'.format(prefix): '',
            '{}-0-col1'.format(prefix): 'v1',
            '{}-0-co'.format(prefix): '1904-106-16',
            '{}-0-ccc'.format(prefix): 'Plain text',
        }
        logger.info("Test get_formset_list()")
        assert get_formset_list(data)

    def test_formset_widget(self):
        """
        Test FormsetdWidget
        """
        choices_string = "col1({'type':'CustomSelectBoxField', 'choices':'v1;v2;v3'})#"\
                         "co({'type':'BaseDateField'})#"\
                         "ccc"
        prefix = "prefix"
        data = {
            '{}-TOTAL_FORMS'.format(prefix): '1',
            '{}-INITIAL_FORMS'.format(prefix): '0',
            '{}-MAX_NUM_FORMS'.format(prefix): '',
            '{}-0-col1'.format(prefix): 'v1',
            '{}-0-co'.format(prefix): '1904-106-16',
            '{}-0-ccc'.format(prefix): 'Plain text',
        }

        formset_field = self.create_field('CustomComplexTableField')[0]
        choices = _split_choices_in_list_canc(choices_string)
        formset_field.widget = FormsetdWidget(field_required=True,
                                              prefix=prefix,
                                              data=data,
                                              files={},
                                              choices=choices)

        logger.info("Test form with formset widget (FormsetdWidget)")
        logger.info(get_formset_labeled_errors(formset_field.widget.formset.errors))
        assert formset_field.widget.get_js_template()
        assert formset_field.widget.make_readonly()



import base64
import re

from django import forms
from django.conf import settings
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from . captcha import get_captcha
from . enc import encrypt, decrypt
from . formsets import build_formset
from . settings import CAPTCHA_DEFAULT_LANG, CAPTCHA_EXPIRATION_TIME, FORMSET_TEMPLATE_NAMEID


FORMSET_TEMPLATE_NAMEID = getattr(settings, 'FORMSET_TEMPLATE_NAMEID',
                                  FORMSET_TEMPLATE_NAMEID)


class CaptchaWidget(forms.Widget):
    template_name = 'widgets/captcha.html'

    def render(self, name, value, attrs=None, renderer=None):
        """Render the widget as an HTML string."""
        context = self.get_context(name, value, attrs)
        captcha_img, captcha_wav = get_captcha(text=self.attrs['value'],
                                               lang=self.attrs.get('lang', getattr(settings,
                                                                                   'CAPTCHA_DEFAULT_LANG',
                                                                                   CAPTCHA_DEFAULT_LANG)))

        # javascript functions don't allow "-" char
        unique_id = name.replace("-", "_")

        # auto-refresh interval time
        expiration_time = getattr(settings, 'CAPTCHA_EXPIRATION_TIME', CAPTCHA_EXPIRATION_TIME)

        # image
        context['image_b64'] = base64.b64encode(captcha_img.read()).decode()
        context['widget']['value'] = ''
        context['widget']['hidden_field'] = self.attrs['hidden_field']
        context['widget']['unique_id'] = unique_id

        # audio
        context['audio_b64'] = base64.b64encode(captcha_wav).decode()

        inline_code = mark_safe( # nosec
            '<script>'
            'var interval = setInterval(refresh_captcha_{unique_id}, {time});'
            'function refresh_captcha_{unique_id}(){{'
                'clearInterval(interval);'
                '$.get(location.href, function(data) {{'
                    '$("#{name}_img").attr("src",($(data).find("#{name}_img").attr("src")));'
                    '$("#{name}_wav").attr("src",($(data).find("#{name}_wav").attr("src")));'
                    '$("#{hidden_field}").val($(data).find("#{hidden_field}").val());'
                '}});'
                'interval = setInterval(refresh_captcha_{unique_id}, {time});'
            '}}'
            '</script>'.format(name=name,
                               unique_id=unique_id,
                               hidden_field=self.attrs['hidden_field'],
                               time=expiration_time)
        )

        return self._render(self.template_name, context, renderer) + inline_code


class FormsetdWidget(forms.Widget):

    def __init__(self, *attrs, **kwargs):
        self.template = 'widgets/formset.html'
        self.readonly = False
        init_data_list = ['choices', 'data', 'files']
        field_required = kwargs.pop('field_required')
        self.prefix = kwargs.pop('prefix')
        for i in init_data_list:
            if i in kwargs.keys():
                setattr(self, i, kwargs.pop(i))
        try:
            if getattr(self, 'data'):
                # It MUST be a list!
                # self.data = [{'ccc': 'dfgdfgdfg', 'co': 'sono', 'col1': 'un', 'data': '2019-04-03'},
                #              {'ccc': '234234', 'co': 'son44o', 'col1': 'un333', 'data': '2016-04-03'}]
                self.formset = build_formset(choices=self.choices,
                                             required=field_required,
                                             prefix=self.prefix,
                                             data=self.data,
                                             files=self.files)
                self.formset.is_valid()
            else:
                # this initialized the formset as void
                self.formset = build_formset(choices=self.choices,
                                             required=field_required,
                                             extra=0,
                                             prefix=self.prefix)
        except:
            # raise Exception('custom message')
            self.formset = None
        super().__init__(*attrs, **kwargs)

    def get_js_template(self):
        formset = build_formset(choices=self.choices,
                                prefix=self.prefix,
                                extra=1)
        res = re.sub('{}-0'.format(self.prefix),
                     '{}-{}'.format(self.prefix,
                                    FORMSET_TEMPLATE_NAMEID),
                     formset.forms[0].as_table())
        return mark_safe(res) # nosec

    def render(self, name='', value='', attrs=None, renderer=None):
        try:
            context_data = {'formset_id': self.prefix,
                            'template_generic_id': FORMSET_TEMPLATE_NAMEID,
                            'formset': self.formset,
                            'formset_template': self.get_js_template(),
                            'readonly': self.readonly,}
            html_output = render_to_string(self.template, context_data)
            return mark_safe(html_output) # nosec
        except Exception as e:
            # raise Exception(e)
            return render_to_string("widgets/formset_exception.html", {'error': e})

    def make_readonly(self, attr='disabled'):
        self.readonly = True
        for form in self.formset:
            for generic_field in form:
                field = form.fields[generic_field.name]
                widget = field.widget
                widget.attrs[attr] = True
        return True



import logging

from django.test import TestCase

from django_form_builder import dynamic_fields
from django_form_builder.forms import BaseDynamicForm
from django_form_builder.models import *
from django_form_builder.settings import ATTACHMENTS_DICT_PREFIX
from django_form_builder.templatetags.django_form_builder_tags import *
from django_form_builder.utils import (get_allegati_dict,
                                       get_as_dict_with_allegati,
                                       get_labeled_errors)

from . base import BaseTest


logger = logging.getLogger('my_logger')

class DynamicFieldMapFake(object):
    """
    Fake generic object to simulate a DynamicFieldMap
    django_form_builder.models.DynamicFieldMap
    """
    pass

class TestFormsets(BaseTest):

    def _simulateDynamicFieldMapEntity(self,
                                       field_type,
                                       name='',
                                       valore=''):
        """
        Simulate a DynamicFieldMap object init
        Using a generic object and inserting needed variables
        """
        field_map_entity = DynamicFieldMapFake()
        field_map_entity.name = name
        field_map_entity.field_type = field_type
        field_map_entity.is_required = True
        field_map_entity.valore = valore
        field_map_entity.aiuto = "example help test"
        return field_map_entity

    def setUp(self):
        """
        Build some field assigning a value to test it
        """
        self.fields = []
        self.data = {}

        name = 'charfield'
        textfield = self._simulateDynamicFieldMapEntity('CustomCharField',
                                                        name=name)
        self.fields.append(textfield)
        self.data[name] = 'testo'

        name = 'integerfield'
        integerfield = self._simulateDynamicFieldMapEntity('PositiveIntegerField',
                                                           name=name)
        self.fields.append(integerfield)
        self.data[name] = 44

        name = 'radiofield'
        radiofield = self._simulateDynamicFieldMapEntity('CustomRadioBoxField',
                                                         name=name,
                                                         valore='col1;col2;col3')
        self.fields.append(radiofield)
        self.data[name] = 'col1'

        name = 'selectfield'
        selectfield = self._simulateDynamicFieldMapEntity('CustomSelectBoxField',
                                                          name=name,
                                                          valore='col1;col2;col3')
        self.fields.append(selectfield)
        self.data[name] = 'col1'

    def test_getform(self):
        """
        Test init form with build_constructor_dict() and fields list
        """
        constructor_dict = BaseDynamicForm.build_constructor_dict(self.fields)
        form = BaseDynamicForm.get_form(constructor_dict=constructor_dict,
                                        data=self.data)
        logger.info("Test form with dynamic fields (build_constructor_dict())")
        assert form.is_valid()

    def test_compiledform(self):
        """
        Test compiled_form() method
        """
        constructor_dict = BaseDynamicForm.build_constructor_dict(self.fields)
        form = SavedFormContent.compiled_form(data_source=json.dumps(self.data))
        logger.info("Test compiledform()")
        assert form.is_valid()

    def test_compiledform_readonly(self):
        """
        Test compiled_form_readonly() method
        """
        constructor_dict = BaseDynamicForm.build_constructor_dict(self.fields)
        form = BaseDynamicForm.get_form(constructor_dict=constructor_dict,
                                        data=self.data)
        SavedFormContent.compiled_form_readonly(form)
        logger.info("Test test_compiledform_readonly()")
        assert form.is_valid()

    def test_not_compiled_field(self):
        """
        Test valid value in PositiveIntegerField
        """
        form = self.single_field_form('PositiveIntegerField', '',
                                      required=False)
        form.remove_not_compiled_fields()
        logger.info("Test remove not compiled fields")
        assert form.is_valid()

    def test_remove_data_field(self):
        """
        Test valid value in PositiveIntegerField
        """
        form = self.single_field_form('PositiveIntegerField', '',
                                      required=False)
        logger.info("Test remove all data fields")
        assert form.is_valid()
        form.remove_datafields()
        assert not form.fields

    def test_remove_filefield(self):
        """
        Test valid value in PositiveIntegerField
        """
        file_test = self.create_fake_file()
        form = self.single_field_form('CustomFileField',
                                      file_value=file_test)
        logger.info("Test valid CustomFileField")
        assert form.is_valid()
        form.remove_files()
        assert not form.fields

    def test_get_dyn_field_name(self):
        """
        Test get_dyn_field_name() method
        """
        field_type = get_dyn_field_name('CustomFileField')
        logger.info("Test get_dyn_field_name()")
        assert field_type

    def test_get_allegati_dict(self):
        """
        Test get_allegati_dict() method
        """
        data = {'data_inizio_dyn': '12/01/2020',
                'data_fine_dyn': '12/03/2020',
                ATTACHMENTS_DICT_PREFIX: {
                    'test_file': 'test_file_path',
                    }
                }
        attachments = get_allegati_dict(compiled_module=data,
                                        pure_text=True)
        logger.info("Test get_allegati_dict()")
        assert attachments

    def test_get_allegati_dict_false(self):
        """
        Test invalid get_allegati_dict() method
        """
        data = {'data_inizio_dyn': '12/01/2020',
                'data_fine_dyn': '12/03/2020'
                }
        attachments = get_allegati_dict(compiled_module=data,
                                        pure_text=True)
        logger.info("Test empty get_allegati_dict()")
        self.assertFalse(attachments)

    def test_get_as_dict_with_allegati(self):
        """
        Test get_as_dict_with_allegati()
        """
        data = {'data_inizio_dyn': '12/01/2020',
                'data_fine_dyn': '12/03/2020',
                ATTACHMENTS_DICT_PREFIX: {
                    'test_file': 'test_file_path',
                    }
                }
        data = get_as_dict_with_allegati(compiled_module=data,
                                         pure_text=True)
        logger.info("Test empty get_allegati_dict()")
        assert data['test_file']



import base64

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from django.conf import settings


_secret = getattr(settings, 'CAPTCHA_SECRET', b'secret')
_salt = getattr(settings, 'CAPTCHA_SALT', b'salt')
_kdf = PBKDF2HMAC(algorithm=hashes.SHA256(),
                  length=32,
                  salt=_salt,
                  iterations=100000,
                  backend=default_backend())

key = base64.urlsafe_b64encode(_kdf.derive(_secret))
_encoder = Fernet(key)

def encrypt(text):
    enc_text = _encoder.encrypt(text.encode())
    return enc_text


def decrypt(b64_text):
    text = base64.b64decode(b64_text)
    decrypted = _encoder.decrypt(text)
    return decrypted



class ChoicesNotDefined(Exception):
    """
    if 'choices' not found in constructor kwargs
    """
    pass


class AudioCaptchaLangPackNotFound(Exception):
    """
    the language pack for Audio CaPTCHA cannot be found
    """
    def __init__(self,
                 msg='the language pack for Audio CaPTCHA cannot be found',
                 *args, **kwargs):
        super().__init__(msg, *args, **kwargs)



import json

from django.db import models
from django.utils.translation import ugettext_lazy as _

from . dynamic_fields import get_fields_types
from . forms import BaseDynamicForm
from . utils import get_as_dict


class DynamicFieldMap(models.Model):
    """
    """
    name = models.CharField(max_length=150,)
    field_type = models.CharField(max_length=100,
                                  choices = get_fields_types())
    valore = models.TextField(max_length=20000,
                              blank=True,
                              default='',
                              verbose_name=_('Lista di Valori'),
                              help_text=_("Viene considerato solo se si sceglie"
                                          " 'Menu a tendina' oppure 'Serie di Opzioni'."
                                          " (Es: valore1;valore2;valore3...)"))
    is_required = models.BooleanField(default=True)
    aiuto = models.CharField(max_length=254, blank=True, default='')
    pre_text = models.TextField(blank=True, default='')
    ordinamento = models.PositiveIntegerField(help_text=_("posizione nell'ordinamento"),
                                              blank=True,
                                              default=0)

    class Meta:
        abstract = True
        ordering = ('ordinamento',)


class SavedFormContent(models.Model):
    """
    """
    # libreria esterna oppure cambio client per JsonField
    # non serve gestirlo come JsonField perchè non vi facciamo ricerche al suo interno ;)
    modulo_compilato = models.TextField()

    @staticmethod
    def compiled_form(data_source=None,
                      constructor_dict={},
                      files=None,
                      remove_filefields=True,
                      remove_datafields=False,
                      form_source=None,
                      # fields_order=[],
                      extra_datas={},
                      **kwargs):
        """
        Returns form compiled by data (json_dict = json.loads(data_source))
        """
        json_dict = json.loads(data_source)
        data = get_as_dict(json_dict, allegati=False)
        if extra_datas:
            for k,v in extra_datas.items():
                data[k]=v
        if not form_source:
            form_source = BaseDynamicForm
        form = form_source.get_form(constructor_dict=constructor_dict,
                                    data=data,
                                    files=files,
                                    remove_filefields=remove_filefields,
                                    remove_datafields=remove_datafields,
                                    **kwargs)
        # Già invocato nel "form_source", ma è bene tenerlo come riferimento
        # if fields_order:
            # form.order_fields(fields_order)
        return form

    @staticmethod
    def compiled_form_readonly(form, attr='disabled', fields_to_remove=[]):
        """
        Returns a more clean version of the compiled_form.
        - Not compiled fields aren't shown (remove_not_compiled_fields());
        - Title attribute by default isn't shown;
        - Form fields are readonly.
        Note: SelectBox aren't affected by readonly attribute!

        This method is useful to produce not editable compiled form.
        """
        form.remove_not_compiled_fields()
        for field_to_remove in fields_to_remove:
            del form.fields[field_to_remove]
        for generic_field in form:
            field = form.fields[generic_field.name]
            widget = field.widget
            widget.attrs[attr] = True
            # If field is a Formset, the widget will make it readonly
            if field.is_formset:
                widget.make_readonly(attr)
                continue
            # Es: TextArea non ha attributo 'input_type'
            # Senza questo controllo il codice genera un'eccezione
            if not hasattr(widget, 'input_type'):
                # widget.attrs[attr] = True
                continue
            tipo = widget.input_type
            if tipo in ['select', 'checkbox', 'radiobox']:
                widget.attrs['disabled'] = True
            # else:
                # widget.attrs[attr] = True
        return form

    class Meta:
        abstract = True



import logging
import os

from django_form_builder import dynamic_fields
from django_form_builder.settings import ATTACHMENTS_DICT_PREFIX, CLASSIFICATION_LIST
from django_form_builder.templatetags.django_form_builder_tags import *
from django_form_builder.utils import format_field_name, get_labeled_errors

from . base import BaseTest

logger = logging.getLogger(__name__)


class TestInvalidFields(BaseTest):

    def test_integer_valid(self):
        """
        Test valid value in PositiveIntegerField
        """
        form = self.single_field_form('PositiveIntegerField', 44)
        logger.info("Test valid PositiveIntegerField")
        assert form.is_valid()

    def test_integer_invalid(self):
        """
        Test invalid value in PositiveIntegerField
        Decimal value instead of Integer
        """
        form = self.single_field_form('PositiveIntegerField', 16e50)
        logger.info("Test invalid PositiveIntegerField")
        logger.info(get_labeled_errors(form))
        self.assertFalse(form.is_valid())

    def test_float_valid(self):
        """
        Test valid value in PositiveFloatField
        """
        form = self.single_field_form('PositiveFloatField', 4)
        logger.info("Test valid PositiveFloatField")
        assert form.is_valid()

    def test_float_invalid(self):
        """
        Test invalid value in PositiveFloatField
        """
        form = self.single_field_form('PositiveFloatField', 16e50)
        logger.info("Test invalid PositiveFloatField")
        logger.info(get_labeled_errors(form))
        self.assertFalse(form.is_valid())

    def test_float_invalid_2(self):
        """
        Test invalid value in PositiveFloatField
        """
        form = self.single_field_form('PositiveFloatField', 4.3333)
        logger.info("Test invalid PositiveFloatField")
        logger.info(get_labeled_errors(form))
        self.assertFalse(form.is_valid())

    def test_date_startend_valid(self):
        """
        Test valid values in DateStartEndComplexField
        """
        data = {'test_field_data_inizio_dyn': '12/01/2020',
                'test_field_data_fine_dyn': '12/03/2020'}
        form = self.complex_field_form('DateStartEndComplexField', data)
        logger.info("Test valid DateStartEndComplexField")
        assert form.is_valid()

    def test_date_startend_invalid(self):
        """
        Test invalid values in DateStartEndComplexField
        Start date > End date
        """
        data = {'test_field_data_inizio_dyn': '12/01/2022',
                'test_field_data_fine_dyn': '12/03/2020'}
        form = self.complex_field_form('DateStartEndComplexField', data)
        logger.info("Test valid DateStartEndComplexField")
        self.assertFalse(form.is_valid())

    def test_ip_valid(self):
        """
        Test invalid values in CustomIPField
        """
        form = self.single_field_form('CustomIPField', '192.168.0.1')
        logger.info("Test valid CustomIPField")
        assert form.is_valid()


    def test_ip_invalid(self):
        """
        Test invalid values in CustomIPField
        """
        form = self.single_field_form('CustomIPField', '192.168000001')
        logger.info("Test invalid CustomIPField")
        self.assertFalse(form.is_valid())

    def test_mac_valid(self):
        """
        Test invalid values in CustomIPField
        """
        form = self.single_field_form('CustomMACField', '98:54:1b:07:15:ed')
        logger.info("Test valid CustomMACField")
        assert form.is_valid()


    def test_mac_invalid(self):
        """
        Test invalid values in CustomIPField
        """
        form = self.single_field_form('CustomMACField', '98.541b:07:15:ed')
        logger.info("Test invalid CustomMACField")
        self.assertFalse(form.is_valid())


    def test_datetimefield_valid(self):
        """
        Test valid values in BaseDateTimeField
        """
        super_label = 'Field label'
        field_id = format_field_name(super_label)
        data = {'{}_data_dyn'.format(field_id): '11/01/2020',
                '{}_ore_dyn'.format(field_id): '12',
                '{}_minuti_dyn'.format(field_id): '12'}
        form = self.complex_field_form('BaseDateTimeField',
                                        data,
                                        super_label)
        logger.info("Test valid BaseDateTimeField")
        assert form.is_valid()

    def test_protocollo_invalid(self):
        """
        Test invalid values in ProtocolloField
        data_protocollo > current date
        """
        data = {'test_field_tipo_numerazione_dyn': CLASSIFICATION_LIST[0][0],
                'test_field_numero_protocollo_delibera_decreto_dyn': '12312 text',
                'test_field_data_protocollo_delibera_decreto_dyn': '12/03/2010'}
        form = self.complex_field_form('ProtocolloField', data)
        logger.info("Test valid ProtocolloField")
        assert form.is_valid()

    def test_filefield(self):
        """
        Test valid value in PositiveIntegerField
        """
        file_test = self.create_fake_file()
        form = self.single_field_form('CustomFileField',
                                      file_value=file_test)
        logger.info("Test valid CustomFileField")
        assert form.is_valid()

    # def test_signed_pdf(self):
        # """
        # Test valid value in PositiveIntegerField
        # """
        # name = 'test_signed.pdf'
        # file_obj = self.get_file(name)
        # form = self.single_field_form('CustomSignedPdfField',
                                      # file_value=file_obj)
        # logger.info("Test valid CustomSignedPdfField")
        # assert form.is_valid()

    # def test_signed_pdf_show_details(self):
        # """
        # Test valid value in PositiveIntegerField
        # """
        # name = 'test_signed.pdf'
        # file_obj = self.get_file(name)
        # field = self.create_field('CustomSignedPdfField')[0]
        # field_id = format_field_name(field.label)
        # initial_fields = {field_id: field}
        # files = {field_id: file_obj}
        # form = self.get_baseform(initial_fields=initial_fields,
                                 # files=files)
        # logger.info("Test valid PDF signed file (sign details)")
        # logger.info(get_attachment_sign_details(form,
                                                # self.get_filepath(),
                                                # field_id,
                                                # name))
        # assert form.is_valid()

    # def test_signed_p7m(self):
        # """
        # Test valid value in PositiveIntegerField
        # """
        # name = 'test_signed.pdf.p7m'
        # file_obj = self.get_file(name, content_type='application/pkcs7-mime')
        # form = self.single_field_form('CustomSignedP7MField', file_value=file_obj)
        # logger.info("Test valid CustomSignedP7MField")
        # assert form.is_valid()



import os
from glob import glob
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

#  rm -R build/ dist/ *egg-info
#  python3 setup.py sdist
#  twine upload dist/*

setup(
    name='django-form-builder',
    version='0.16.10',
    packages=find_packages(),
    package_data={'': ['*.wav']},
    data_files=[
        ('', glob('django_form_builder/data/audio_captcha/*/default.wav', recursive=True)),
    ],
    include_package_data=True,
    license='BSD License',
    description="Django Form builder",
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/UniversitaDellaCalabria/django-form-builder',
    author='Giuseppe De Marco, Francesco Filicetti',
    author_email='giuseppe.demarco@unical.it, francesco.filicetti@unical.it',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        "License :: OSI Approved :: Apache Software License",
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    install_requires=[
        'django>=2.0,<4.0',
        'filesig>=0.3',
        'cryptography>=2.8',
        'captcha==0.3'
    ],
)



import copy
import re

from collections import OrderedDict
from django import forms
from django.conf import settings
from django.forms.fields import FileField
from django.http import QueryDict
from django.template.defaultfilters import filesizeformat
from django.utils.module_loading import import_string

from . import dynamic_fields
from . settings import FORMSET_REGEX
from . utils import format_field_name, _split_choices_in_list_canc
from . widgets import FormsetdWidget


def _remove_filefields_from_formset(formset):
    for f_form in formset.forms:
        to_be_removed_formset = []
        for f_field in f_form.fields:
            if not isinstance(f_form.fields[f_field], FileField):
                to_be_removed_formset.append(f_field)
        for i in to_be_removed_formset:
            del f_form.fields[i]


class BaseDynamicForm(forms.Form):
    def __init__(self,
                 fields_source=dynamic_fields,
                 initial_fields={},
                 final_fields={},
                 constructor_dict={},
                 custom_params={},
                 ignore_format_field_name=False,
                 *args,
                 **kwargs):
        """
        Base class for form building.
        User custom forms have to inherit from this.
        Example: class MyCustomFormClass(BaseDynamicForm)

        :type fields_source: String
        :type initial_fields: dict
        :type final_fields: dict
        :type constructor_dict: OrderedDict
        :type custom_params: dict
        :type ignore_format_field_name: boolean

        :param fields_source: fields source file (default: dynamic_fields.py, but you can create your own to extend dynamic_fields.py)
        :param initial_fields: static initial fields to render in every instance of your form
        :param final_fields: static final fields to render in every instance of your form
        :param constructor_dict: form fields to render between initial and final fields
        :param custom_params: extra params you may need your own form
        :param ignore_format_field_name: choose if build or not fields names from labels (dynamic_fields.format_field_name() method)

        :example initial_fields:
        # Generate static NAME field
        name_id = format_field_name("Personal name")
        # Set NAME field data
        name_data = {'required' : True,
                     'label': "Name",
                     'help_text': "Your help-text"}
        # Build NAME field (define type and pass arguments)
        name_field = getattr(dynamic_fields, 'CustomCharField')(**name_data)
        # Put NAME field in initial_fields{}
        self.initial_fields[name_id] = name_field

        :example final_fields:
        same as initial_fields, but use initial_fields{} dict

        :example constructor_dict
        constructor_dict = OrderedDict([
            # CharField
            ('Phone',
                ('CustomCharField',
                    {'label': 'Phone number',
                     'required': True,
                     'help_text': 'Mobile',
                     'pre_text': ''},
                    '')
            ),
            # Complex field (start date / end date)
            ('From to',
                ('DateStartEndComplexField',
                    {'label': 'From to',
                     'required': True,
                     'help_text': 'From date to date',
                     'pre_text': ''},
                    '')
            ),
            # Formset
            ('Formset data',
                ('CustomComplexTableField',
                    {'label': 'Formset data',
                     'required': True,
                     'help_text': '',
                     'pre_text': 'This is a formset, this text is printed before rendering field'},
                    # Columns of tables with different field types
                    'select_option({"type":"CustomSelectBoxField","choices":"v1;v2;v3"})'
                    '#'
                    'simple_text'
                    '#'
                    'email({"type":"CustomEmailField"})'
                    '#'
                    'phone({"type":"PositiveIntegerField"})'
                    '#'
                    'valid_until({"type":"BaseDateField"})')
            ),
            # Captcha
            ('CaPTCHA',
                ('CustomCaptchaComplexField',
                    {'label': 'CaPTCHA',
                     'pre_text': ''},
                    '')
            ),
        ])

        :example custom_params:
        custom_params = {'my_custom_param': True,
                         'another_param_to_pass_to_form': 'value',}
        """
        super().__init__(*args, **kwargs)

        # if initial fields are present
        # form fields start from them
        self.fields = initial_fields or self.fields

        # if there's a constructor_dict
        # start fields building
        if constructor_dict:

            # make a deepcopy avoiding to edit dict passed as argument
            constructor_dict = copy.deepcopy(constructor_dict)

            # for every key in dict, start building field
            for key, value in constructor_dict.items():

                # if ignore_format_field_name
                # then field_id is the key
                if ignore_format_field_name:
                    field_id = key
                # else build id with utility method
                else:
                    field_id = format_field_name(key)

                # start building field data kwargs
                # default contructor data (capitalize key for label)
                data_kwargs = {'label': key.title()}

                # see constructor_dict example in docstring
                # This is constructor_dict element value for key 'Phone'
                # ('CustomCharField',
                #  {'label': 'Phone number',
                #   'required': True,
                #   'help_text': 'Mobile',
                #   'pre_text': ''},
                # '')

                # value[0] is field_type (CustomCharField in this case)
                custom_field_name = value[0]

                # value[1] is field data dict (label, required, etc...)
                custom_field_dict = value[1]

                # value[2] is field value (if needed from field.)
                # Empty in this example case
                custom_field_values = value[2]

                # get field pre_text (if key exists) and pop from dict
                pre_text = ''
                if 'pre_text' in custom_field_dict.keys():
                    pre_text = custom_field_dict.pop('pre_text')

                # update field data kwargs
                data_kwargs.update(custom_field_dict)

                # get field object from fields_source
                custom_field = self._get_object_from_fields_source(fields_source,
                                                                   custom_field_name,
                                                                   data_kwargs)

                # if field has been created
                if custom_field:
                    # call define_value method to set field value
                    # and pass custom_params dict if it needs
                    custom_field.define_value(custom_field_values,
                                              **custom_params)

                    # if custom_field is a complex field returns a list with two or more elements.
                    # if it's a simple field returns a list with one element
                    fields = custom_field.get_fields()

                    # for every field in custom field
                    for field in fields:
                        # set name
                        name = getattr(field, 'name', field_id)

                        # add field to form fields (self.fields)
                        self.fields[name]= field

                        # if pre_text is setted add it to form field
                        if pre_text:
                            self.fields[name].pre_text = pre_text

                        # if is a Formset
                        if isinstance(field,
                                      dynamic_fields.CustomComplexTableField):

                            # get formset choices with the utility method
                            # _split_choices_in_list_canc()
                            # (pass custom_field_values string as argument, see above)
                            choices = _split_choices_in_list_canc(custom_field_values)

                            # add formset widget to form field
                            self.fields[name].widget = FormsetdWidget(choices=choices,
                                                                      data=kwargs.get('data',{}),
                                                                      files=kwargs.get('files',{}),
                                                                      field_required=field.required,
                                                                      prefix=name)
                        # else if field isn't a Formset
                        else:
                            # get a custom widget (if user has defined it in its settings file)
                            custom_widget = getattr(settings, 'CUSTOM_WIDGETS', {}).get(field.__class__.__name__)

                            # if theresn't a custom widget, then apply the standard field widget
                            if not custom_widget:
                                custom_widget = self.fields[name].widget

                            # Manage widget declaration
                            # (class or string? manage both)
                            if isinstance(custom_widget, str):
                                self.fields[name].widget = import_string(custom_widget)()
                            else:
                                self.fields[name].widget = custom_widget

                            # set form field choices attribute
                            if hasattr(field, 'choices'):
                                self.fields[name].choices = getattr(field, 'choices')

        # if final fields are present append them to form
        self.fields.update(final_fields)

    def _get_object_from_fields_source(self,
                                       fields_source,
                                       field_name,
                                       data):
        """
        if fields_source defines custom_field_name class,
        get an instance of that class ad pass data to it
        """
        if not fields_source: return False
        if not field_name: return False
        if hasattr(fields_source, field_name):
            return getattr(fields_source, field_name)(**data)
        return False

    def remove_not_compiled_fields(self):
        """
        Remove all not filled fields from a compiled form.
        Used in get_form() method
        """
        to_be_removed = []
        for field in self:
            form_field = self.fields[field.name]

            # remove not compiled single fields in formset
            if form_field.is_formset:
                formset = form_field.widget.formset
                for f_form in formset:
                    to_be_removed_formset = []
                    for generic_field in f_form:
                       if not generic_field.value():
                            to_be_removed_formset.append(generic_field.name)
                    for i in to_be_removed_formset:
                        del f_form.fields[i]
                continue

            if not field.value():
                to_be_removed.append(field.name)
        for i in to_be_removed:
            del self.fields[i]

    def remove_datafields(self):
        """
        Remove all fields different from FileField
        """
        to_be_removed = []
        for field in self.fields:
            if isinstance(self.fields[field], dynamic_fields.CustomComplexTableField):
                _remove_filefields_from_formset(self.fields[field].widget.formset)
            elif not isinstance(self.fields[field], FileField):
                to_be_removed.append(field)
        for i in to_be_removed:
            del self.fields[i]

    def remove_files(self, allegati=None):
        """
        Remove all FileFields.
        Used in edit form if attachments are present
        """
        REGEX = getattr(settings, 'FORMSET_REGEX', FORMSET_REGEX)
        to_be_removed = []
        for field in self.fields:
            if isinstance(allegati, dict):
                for allegato in allegati:
                    if field==allegato:
                        to_be_removed.append(field)
                    else:
                        formset_regex = re.match(FORMSET_REGEX.format(field), allegato)
                        if formset_regex:
                            index = int(formset_regex['index'])
                            name = formset_regex['name']
                            formset = self.fields[field].widget.formset
                            try:
                                del formset.forms[index].fields[name]
                            except Exception: continue # nosec
            else:
                # remove all attachments/files
                if isinstance(self.fields[field], FileField):
                    to_be_removed.append(field)
                elif isinstance(self.fields[field], dynamic_fields.CustomComplexTableField):
                    _remove_filefields_from_formset(self.fields[field].widget.formset)
        for i in to_be_removed:
            del self.fields[i]

    def clean(self, *args, **kwargs):
        cleaned_data = super().clean()

        self.data = copy.deepcopy(self.data)

        # CaPTCHA MUST BE ALWAYS RENEWED!
        for field_name, field_obj in self.fields.items():
            if type(field_obj) in (dynamic_fields.CaptchaField,
                                   dynamic_fields.CaptchaHiddenField):
                self.data[field_name] = self.fields[field_name].widget.attrs['value']
        # end CAPTCHA

        for fname in self.fields:
            field = self.fields[fname]

            # formset is empty or not valid
            if field.is_formset:
                if not field.widget.formset:
                    continue
                for f_form in field.widget.formset.forms:
                    for f_field_name in f_form.fields:
                        f_field = f_form.fields[f_field_name]
                        errors = f_field.raise_error(f_field_name,
                                                     f_form.cleaned_data.get(f_field_name),
                                                     **kwargs)
                        if errors:
                            f_form.add_error(f_field_name, errors)
                    if not f_form.is_valid():
                        f_form_errors = f_form.errors.as_data()
                        for error_field in f_form_errors:
                            if error_field in f_form.fields:
                                self.add_error(fname, f_form_errors[error_field])

            # other fields check
            # if field is a child of a complex field
            if hasattr(field, 'parent'):
                field = getattr(field, 'parent')
                errors = field.raise_error(fname,
                                           cleaned_data,
                                           **kwargs)
            # else if is a simple field
            else:
                errors = field.raise_error(None,
                                           cleaned_data.get(fname),
                                           **kwargs)

            # if errors are present
            if errors:
                self.add_error(fname, errors)
                continue

    @staticmethod
    def build_constructor_dict(fields):
        """
        Static method that builds the constructor_dict OrderedDict
        from a list/queryset of fields.
        Useful when fields are managed by RDBMS
        """
        constructor_dict = OrderedDict()
        for field in fields:
            field_params = {'label': field.name,
                            'required' : field.is_required,
                            'help_text' : field.aiuto,
                            'pre_text': getattr(field, 'pre_text', '')}
            constructor_dict[field.name] = (field.field_type,
                                            field_params,
                                            field.valore)
        return constructor_dict

    @classmethod
    def get_form(cls,
                 class_obj=None,
                 constructor_dict={},
                 custom_params={},
                 remove_filefields=False,
                 remove_datafields=False,
                 fields_order=[],
                 *args, **kwargs):
        """
        Static method that builds a form from a constructor_dict,
        custom_params (if present) and a source class (class_obj)
        """
        # form = class_obj(constructor_dict=constructor_dict,
                         # custom_params={} if class_obj == BaseDynamicForm else custom_params,
                         # *args, **kwargs)
        constructor_class = class_obj or cls
        form = constructor_class(constructor_dict=constructor_dict,
                                 custom_params=custom_params,
                                 *args, **kwargs)

        if remove_filefields:
            form.remove_files(allegati=remove_filefields)
        if remove_datafields:
            form.remove_datafields()
        if fields_order:
            form.order_fields(fields_order)
        return form

    class Media:
        # javascript to formset dynamic management
        js = ('js/formset_js.js',)



import os
import random
import string

from captcha.audio import AudioCaptcha
from captcha.image import ImageCaptcha

from django.conf import settings
from . exceptions import AudioCaptchaLangPackNotFound
from . settings import CAPTCHA_DEFAULT_LANG


def get_captcha(text=None, lang=getattr(settings,
                                        'CAPTCHA_DEFAULT_LANG',
                                        CAPTCHA_DEFAULT_LANG)):
    fonts = getattr(settings, 'CAPTCHA_FONTS', None)
    length = getattr(settings, 'CAPTCHA_LENGTH', 5)
    voicedir = getattr(settings, 'CAPTCHA_VOICEDIR',
                       os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                    'data', 'audio_captcha', lang))
    # if audio for that language is not available
    # fallback to settings default (english: en)
    if not os.path.exists(voicedir):
        voicedir = getattr(settings, 'CAPTCHA_VOICEDIR',
                           os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                        'data', 'audio_captcha',
                                        CAPTCHA_DEFAULT_LANG))
    image = ImageCaptcha(fonts=fonts)

    try:
        audio = AudioCaptcha(voicedir=voicedir)
    except IndexError as e:
        raise AudioCaptchaLangPackNotFound()

    text = text or ''.join([random.choice(string.ascii_letters+string.hexdigits) # nosec
                            for i in range(length)])

    data_image = image.generate(text)
    data_audio = audio.generate(text)
    return data_image, data_audio
